__all__ = (
    'MainCorpus',
    'DialectalCorpus',
    'AccentologicalCorpus',
    'SpokenCorpus',
    'PaperRegionalCorpus',
    'Paper2000Corpus',
    'ParallelCorpus',
    'MultilingualParaCorpus',
    'TutoringCorpus',
    'MultimodalCorpus',

    'SORT_KEYS',
    'SEARCH_FORMATS',
    'OUTPUT_FORMATS'
)

import asyncio
import csv
import logging
import os
import random
import re
import string
import time
import urllib.parse
from abc import ABC, abstractmethod
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Dict, Callable, List, Any, Tuple, Pattern

import bs4
import ujson

import rnc.corpora_requests as creq
import rnc.examples as expl

logger = logging.getLogger("rnc")


# Russian National Corpus URL
RNC_URL = "https://processing.ruscorpora.ru/search.xml"
BASE_RNC_URL = "https://processing.ruscorpora.ru"

ALPHABET = f"{string.ascii_letters}{string.digits}"


SORT_KEYS = (
    'i_grtagging', 'random', 'i_grauthor', 'i_grcreated_inv',
    'i_grcreated', 'i_grbirthday_inv', 'i_grbirthday',
)
SEARCH_FORMATS = (
    'lexform', 'lexgramm'
)
OUTPUT_FORMATS = (
    'normal', 'kwic'
)


def create_filename(length: int = 8) -> str:
    """ Create random filename. """
    name = random.sample(ALPHABET, length)
    return ''.join(name)


def create_unique_filename(folder: Path,
                           class_name: str,
                           p_count: int) -> Path:
    """ Create a random unique csv filename,
    means the file does not exist.

    Name format:
    <class_name><p_count>_<some random symbols>
    """
    name_template = "{}{}_{}.csv"
    name = name_template.format(class_name, p_count, create_filename())
    path = folder / name
    while path.exists():
        name = name_template.format(class_name, p_count, create_filename())
        path = path.with_name(name)
    return path


def clean_text_up(text: str) -> str:
    """ Remove duplicate spaces from str and strip it. """
    return ' '.join(text.split()).strip()


def create_doc_url(doc_url: str) -> str:
    """ Create full url to document in RNC. Add https://... to the url. """
    if not doc_url:
        return doc_url
    return f"{BASE_RNC_URL}/{doc_url}"


def join_with_plus(item: str) -> str:
    """ Split txt and join it with '+'. """
    res = item.split()
    return '+'.join(res)


def str_to_int(value: str) -> int:
    """ Convert str like '350 000 134' to int. """
    return int(value.replace(' ', ''))


class Corpus(ABC):
    """ Base class for Corpora """
    # default params
    # documents per page
    __DPP = 5
    # sentences per document
    __SPD = 10
    # search format
    __TEXT = 'lexgramm'
    # output format
    __OUT = 'normal'
    # accent on words
    __ACCENT = '0'
    # show order
    __SORT = 'i_grtagging'
    # distance between n and (n + 1) words
    __MIN = 1
    __MAX = 3
    # count of examples to print
    __RESTRICT_SHOW = 50

    # symbols to write csv
    _DATA_W_DELIMITER = '\t'
    _DATA_W_QUOTCHAR = '"'

    DATA_FOLDER = Path('data')

    def __init__(self,
                 query: dict or str = None,
                 p_count: int = None,
                 file: str or Path = None,
                 **kwargs) -> None:
        """ 
        If the file exists, working with a local database.

        :param query: dict of str, words to search;
         {word1: {properties}, word2: {properties}...}.
         If you chose 'lexform' as a 'text' param, you must give here a string.
        :param p_count: int, count of pages to request.
        :param file: str or Path, filename of a local database.
         Optional, random filename by default.
        :keyword dpp: str or int, documents per page.
         Optional, 5 by default.
        :keyword spd: str or int, sentences per document.
         Optional, 10 by default.
        :keyword text: str, search format: 'lexgramm' or 'lexform'.
         Optional, 'lexgramm' by default.
        :keyword out: str, output format: 'normal' or 'kwic'.
         Optional, 'normal' bu default.
        :keyword kwsz: str or int, count of words in context;
         Optional param if 'out' is 'kwic'.
        :keyword sort: str, sort show order. See docs how to set it.
         Optional.
        :keyword mycorp: str, mycorp. This is way to specify the sample of docs
         where you want to find sth. See docs how to set it. Optional.
        :keyword expand: str, if 'full', all part of doc will be shown. 
         Now it doesn't work.
        :keyword accent: str or int, with accents on words or not:
         1 – with, 0 – without. Optional, 0 by default.
        :keyword marker: function, with which found words will be marked.
         Optional.

        :exception FileExistsError: if csv file is given but json file
         with config doesn't exist.
        :exception ValueError: if the query is empty; page count is a negative 
         number; text, out or sort key is wrong.
        :exception NotImplementedError: if the corpus type in file isn't equal 
         to corpus class type.
        """
        # list of examples
        self._data = []
        # http tags to request
        self._params = {}
        # found wordforms with their frequency
        self._found_wordforms = defaultdict(int)
        # query, wordforms to find
        self._query = {}
        # count of PAGES
        self._p_count = 0
        # type of example should be defined before params init
        self._ex_type = kwargs.pop('ex_type', None)
        self._marker = kwargs.pop('marker', None)
        # additional info from the first page:
        # amount of docs, contexts, where the query was found,
        # link to the graphic with distribution by years
        self._add_info = {}

        # path to local database
        class_name = self.__class__.__name__.replace('Corpus', '')
        path = file or create_unique_filename(
            self.DATA_FOLDER, class_name, p_count)
        path = Path(path)
        # change or add right extension
        path = path.with_suffix('.csv')

        # to these files the data and req params will be dumped
        self._csv_path = path
        self._config_path = path.with_suffix('.json')

        # init from file if it exists
        if self._csv_path.exists():
            try:
                self._from_file()
            except FileExistsError as e:
                logger.error(
                    f"Config file not found: '{self._config_path}'\n{e}")
                raise
        # or work with RNC
        else:
            self._from_corpus(query, p_count, **kwargs)

    def _from_corpus(self,
                     query: dict or str,
                     p_count: int,
                     **kwargs) -> None:
        """ Set given values to the object. If the file does not exist.
        Params the same as in the init method.

        :exception ValueError: if the query is empty; pages count is a negative
         number; out, sort, text key is wrong.
        """
        if not query:
            msg = f"Query must be not empty, but '{query}' found"
            logger.error(msg)
            raise ValueError(msg)
        self._query = query

        if p_count <= 0:
            msg = f"Pages count must be > 0, but '{p_count}' found"
            logger.error(msg)
            raise ValueError(msg)
        self._p_count = p_count

        # base params
        self._params['env'] = 'alpha'
        self._params['api'] = '1.0'
        self._params['lang'] = 'en'

        self._params['dpp'] = kwargs.pop('dpp', None) or self.__DPP
        self._params['spd'] = kwargs.pop('spd', None) or self.__SPD
        self._params['text'] = kwargs.pop('text', None) or self.__TEXT
        self._params['out'] = kwargs.pop('out', None) or self.__OUT
        self._params['sort'] = kwargs.pop('sort', None) or self.__SORT

        msg = "'{}' is wrong {} value, expected: {}"
        if self.params['text'] not in SEARCH_FORMATS:
            msg = msg.format(self.params['text'], 'text', SEARCH_FORMATS)
            logger.error(msg)
            raise ValueError(msg)
        if self.params['out'] not in OUTPUT_FORMATS:
            msg = msg.format(self.params['out'], 'out', OUTPUT_FORMATS)
            logger.error(msg)
            raise ValueError(msg)
        if self.params['sort'] not in SORT_KEYS:
            msg = msg.format(self.params['sort'], 'sort', SORT_KEYS)
            logger.error(msg)
            raise ValueError(msg)

        accent = kwargs.pop('accent', None) or self.__ACCENT
        accent = int(accent)
        self._params['nodia'] = int(not accent)

        if self.params['out'] == 'kwic' and 'kwsz' in kwargs:
            self._params['kwsz'] = kwargs.pop('kwsz')

        if 'mycorp' in kwargs:
            mycorp = kwargs.pop('mycorp')
            self._params['mycorp'] = urllib.parse.unquote(mycorp)

        # TODO: page structure changed if expand=full
        # if 'expand' in kwargs:
        #     self._params['expand'] = kwargs.pop('expand')

        self._query_to_http()

        # parsing depends on 'out' value
        self._page_parser = None
        self._page_parser_and_ex_type()

    def _from_file(self) -> None:
        """ Load data and params from the local databases.
        If the file exists.

        :exception FileExistsError: if csv file with data or
         json file with config do not exist.
        """
        if not (self._csv_path.exists() and self._config_path.exists()):
            raise FileExistsError("Data and config file must exist together")

        params = self._load_params()
        self._query = params.get('query', None)
        self._p_count = params.get('p_count', None)
        self._params = params.get('params', None)

        mode = self.mode
        if mode is None or mode != self._MODE:
            msg = f"Tried to load data to wrong Corpus: " \
                  f"{mode} instead of {self._MODE}"
            logger.error(msg)
            raise NotImplementedError(msg)

        # these params must be defined here too
        self._page_parser_and_ex_type()

        self._data = self._load_data()
        # add info about
        try:
            self._get_additional_info()
        except Exception as e:
            logger.warning("It is impossible to get "
                           f"additional info from RNC:\n{e}")

    def _load_data(self) -> List:
        """ Load data from csv file. """
        with self.file.open('r', encoding='utf-8') as f:
            dm = self._DATA_W_DELIMITER
            qch = self._DATA_W_QUOTCHAR
            reader = csv.reader(f, delimiter=dm, quotechar=qch)
            # first row contains headers, skip it
            next(reader)

            data = [self.ex_type(*row) for row in reader]
            wordforms = list(map(lambda example: example.found_wordforms, data))
            wordforms = sum(wordforms, [])
            self._add_wordforms(wordforms)

        return data

    def _load_params(self) -> Dict:
        """ Load request params from json file. """
        with self._config_path.open('r', encoding='utf-8') as f:
            return ujson.load(f)

    @classmethod
    def set_dpp(cls, value: int) -> None:
        if not isinstance(value, int) or value <= 0:
            msg = f"DPP must be int > 0, but '{value}' found"
            logger.error(msg)
            raise ValueError(msg)
        cls.__DPP = value

    @classmethod
    def set_spd(cls, value: int) -> None:
        if not isinstance(value, int) or value <= 0:
            msg = f"SPD must be int > 0, but '{value}' found"
            logger.error(msg)
            raise ValueError(msg)
        cls.__SPD = value

    @classmethod
    def set_text(cls, value: str) -> None:
        if value not in SEARCH_FORMATS:
            msg = f"Search formats: {SEARCH_FORMATS},\nbut '{value}' found"
            logger.error(msg)
            raise ValueError(msg)

        cls.__TEXT = value

    @classmethod
    def set_sort(cls, value: str) -> None:
        if value not in SORT_KEYS:
            msg = f"Sort keys: {SORT_KEYS},\nbut '{value}' found"
            logger.error(msg)
            raise ValueError(msg)

        cls.__SORT = value

    @classmethod
    def set_out(cls, value: str) -> None:
        if value not in OUTPUT_FORMATS:
            msg = f"Out formats: {OUTPUT_FORMATS},\nbut '{value}' found"
            logger.error(msg)
            raise ValueError(msg)

        cls.__OUT = value

    @classmethod
    def set_min(cls, value: int) -> None:
        if not isinstance(value, int) or value <= 0:
            msg = f"min must be int > 0, but '{value}' found"
            logger.error(msg)
            raise ValueError(msg)
        cls.__MIN = value

    @classmethod
    def set_max(cls, value: int) -> None:
        if not isinstance(value, int) or value <= 0:
            msg = f"max must be int > 0, but '{value}' found"
            logger.error(msg)
            raise ValueError(msg)
        cls.__MAX = value

    @classmethod
    def set_restrict_show(cls, value: int or bool) -> None:
        """ Set amount of showing examples.
        Show all examples if `False` given.
        """
        if not isinstance(value, (int, bool)):
            msg = f"Restrict count must be int or bool, but '{value}' found"
            logger.error(msg)
            raise TypeError(msg)
        cls.__RESTRICT_SHOW = value

    @staticmethod
    def _get_ambiguation(tag: bs4.element.Tag) -> str:
        """ Get pretty ambiguation from example.

        :return: 'disambiguated' or 'not disambiguated' or 'Not found'.
        """
        ambiguation = (tag.find('span', {'class': 'off'}) or
                       tag.find('span', {'class': 'on'}))
        if not ambiguation:
            return 'Not found'
        ambiguation = ambiguation.text.strip()

        # TODO: use regexp here
        # here ambiguation like '[...]'
        ambiguation = ambiguation[1:-1].strip()
        return ambiguation

    @staticmethod
    def _get_text(tag: bs4.element.Tag) -> str:
        """ Get pretty text from example and remove
        from there duplicate spaces.

        Here it is assumed, that all examples have text.
        """
        # using 'findall' method removes punctuation marks
        txt = tag.get_text()
        # remove duplicate spaces
        return clean_text_up(txt)

    @staticmethod
    def _get_doc_url(tag: bs4.element.Tag) -> str:
        """ Get pretty doc url from example.

        :return: doc url or 'Not found'.
        """
        doc_url = tag.a
        if not doc_url:
            return 'Not found'
        doc_url = doc_url.attrs['href']
        return create_doc_url(doc_url)

    @staticmethod
    def _get_source(tag: bs4.element.Tag) -> str:
        """ Get pretty source from example.

        :return: examples source or 'Not found'.
        """
        src = tag.find('span', {'class': 'doc'})
        if not src:
            return "Not found"
        src = clean_text_up(src.text)

        # TODO: use regexp gere
        # here src like '[...]'
        src = src[1:-1].strip()
        return src

    @staticmethod
    def _parse_lexgramm_params(params: dict or str,
                               join_inside_symbol: str,
                               with_braces: bool = False) -> str:
        """ Convert lexgramm params to str for HTTP request.

        :param join_inside_symbol: symbol to join params.
        :param with_braces: whether the braces
         will be added around the param.
        :return: joined with ',' params.
        """
        if not (isinstance(params, (str, dict)) and ' ' not in params):
            msg = f"Params must be str without spaces " \
                  f"or dict, but '{params}' found"
            logger.error(msg)
            raise ValueError(msg)
        # let the user to give only one param:
        # {'word': {'gramm': 'acc', 'flags': 'bmark'}}
        if isinstance(params, str):
            return params

        res = []
        for val in params.values():
            if not isinstance(val, (str, int, list)):
                msg = f"One should give to tags only " \
                      f"str, list or int, but '{val}' found"
                logger.error(msg)
                raise ValueError(msg)

            if isinstance(val, (str, int)):
                val = [str(val)]

            item = f"{'(' * with_braces}" \
                   f"{join_inside_symbol.join(val)}" \
                   f"{')' * with_braces}"
            res += [item]
        return ','.join(res)

    @staticmethod
    def _find_searched_words(tag: bs4.element.Tag) -> List[str]:
        """ Get found words, they are marked with 'g-em'
        parameter in the class name. Strip them.
        """
        # searched words are marked by class parameter 'g-em'
        return [
            tag.text.strip()
            for tag in tag.contents
            if (isinstance(tag, bs4.element.Tag) and
                'g-em' in tag.attrs.get('class', ''))
        ]

    @property
    def data(self) -> List:
        """ Get list of all examples """
        return self._data

    @property
    def query(self) -> Dict[str, dict] or str:
        """ Get requested words items (dict of words
        with params or str with words)
        """
        return self._query

    @property
    def forms_in_query(self) -> List[str]:
        """ Requested wordforms """
        req = self.query
        if isinstance(req, str):
            return req.split()
        return list(req.keys())

    @property
    def p_count(self) -> int:
        """ Requested count of pages """
        return self._p_count

    @property
    def file(self) -> Path:
        """ Get path to local database file. """
        return self._csv_path

    @property
    def marker(self) -> Callable:
        """ Get function to mark found wordforms. """
        return self._marker

    @property
    def params(self) -> dict:
        """ Get all HTTP params """
        return self._params

    @property
    def found_wordforms(self) -> Dict[str, int]:
        """ Get info about found wordforms, {form: frequency}. """
        return self._found_wordforms

    @property
    def url(self) -> str:
        """ Get URL to first page of RNC results. """
        params = '&'.join(
            f"{key}={val}"
            for key, val in self.params.items()
        )
        return f"{RNC_URL}?{params}"

    @property
    def ex_type(self) -> Any:
        """ get type of Example objects """
        return self._ex_type

    @property
    def amount_of_docs(self) -> int or None:
        """ Get amount of documents, where the query was found
        or None if there's no this info.
        """
        return self._add_info.get('docs', None)

    @property
    def amount_of_contexts(self) -> int or None:
        """ Get amount of contexts, where the query was found
        or None if there's no this info.
        """
        return self._add_info.get('contexts', None)

    @property
    def graphic_link(self) -> str or None:
        """ Get the link to the graphic
        or None if there's no this info.
        """
        return self._add_info.get('graphic_link', None)

    @staticmethod
    def _get_where_query_found(content: bs4.element.Tag) -> Dict[str, Any]:
        """ Get converted to int amount of found docs and contexts. """
        res = {}
        amount = list(content.find_all('p', {'class': 'res'}))
        blocks = amount[-1].find_all('span', {'class': 'stat-number'})

        contexts = blocks[-1].get_text()
        res['contexts'] = str_to_int(contexts)
        if len(blocks) == 2:
            docs = blocks[0].get_text()
            res['docs'] = str_to_int(docs)
        return res

    @staticmethod
    def _get_graphic_url(content: bs4.element.Tag) -> str or None:
        """ Get URL to the graphic. """
        a = content.find('a', {'target': '_blank'})
        try:
            link = a['href']
        except (KeyError, TypeError, AttributeError):
            return
        return f"{BASE_RNC_URL}/{link}"

    def _get_additional_info(self,
                             first_page: str = None) -> None:
        """ Get additional info (amount of found
        docs and contexts, link to the graphic).
        """
        params = self.params.copy()
        params['lang'] = 'ru'
        params.pop('expand', None)
        try:
            first_page = first_page or creq.get_htmls(RNC_URL, **params)[0]
        except creq.BaseRequestError:
            raise

        soup = bs4.BeautifulSoup(first_page, 'lxml')
        content = soup.find('div', {'class': 'content'})

        try:
            additional_info = Corpus._get_where_query_found(content)
            graphic_url = Corpus._get_graphic_url(content)
        except Exception as e:
            logger.error("Sth went wrong while "
                         f"getting additional info:\n{e}")
        else:
            if graphic_url:
                additional_info['graphic_link'] = graphic_url

            self._add_info = additional_info

    async def _get_additional_info_async(self,
                                         first_page: str = None) -> None:
        """ Get additional info (amount of found
        docs and contexts, link to the graphic).
        """
        params = self.params.copy()
        params['lang'] = 'ru'
        params.pop('expand', None)
        try:
            first_page = first_page or \
                         (await creq.get_htmls_async(RNC_URL, **params))[0]
        except creq.BaseRequestError:
            raise

        soup = bs4.BeautifulSoup(first_page, 'lxml')
        content = soup.find('div', {'class': 'content'})

        try:
            additional_info = Corpus._get_where_query_found(content)
            graphic_url = Corpus._get_graphic_url(content)
        except Exception as e:
            logger.error("Sth went wrong while "
                         f"getting additional info:\n{e}")
        else:
            if graphic_url:
                additional_info['graphic_link'] = graphic_url

            self._add_info = additional_info

    def _page_parser_and_ex_type(self) -> None:
        """ Add 'parser' and 'ex_type' params.
        They are depended on 'out' tag.
        """
        if self.out == 'normal':
            # ex_type is defined above in this case
            self._page_parser = self._parse_page_normal
        elif self.out == 'kwic':
            self._page_parser = self._parse_page_kwic
            self._ex_type = expl.KwicExample

    def _query_to_http(self) -> None:
        """ Convert the query to HTTP tags, add them to params.

        :exception ValueError: if the query is not
        str however out is lexform;
        """
        if self.text == 'lexform':
            if not isinstance(self.query, str):
                msg = f"Query must be str if search is " \
                      f"'lexform', but '{self.query}' found"
                logger.error(msg)
                raise ValueError(msg)
            self._params['req'] = join_with_plus(self.query)

        # in lexgramm search one word may be too
        if isinstance(self.query, str):
            # working with long query like
            # 'открыть -открыл дверь -двери настеж'
            q = re.finditer(r'\b\w+\b( -\b\w+\b)?', self.query)
            for num, query in enumerate(q, 1):
                match = query.group(0)
                self._params[f"lex{num}"] = join_with_plus(match)
                if num > 1:
                    self._params[f"min{num}"] = self.__MIN
                    self._params[f"max{num}"] = self.__MAX
            return

        # words and their params
        for word_num, (word, params) in enumerate(self.query.items(), 1):
            # add distance
            if word_num > 1:
                min_ = f'min{word_num}'
                max_ = f'max{word_num}'
                # given or default values
                if isinstance(params, dict):
                    self._params[min_] = params.get('min', None) or self.__MIN
                    self._params[max_] = params.get('max', None) or self.__MAX
                else:
                    self._params[min_] = self.__MIN
                    self._params[max_] = self.__MAX

            self._params[f"lex{word_num}"] = join_with_plus(word)

            if isinstance(params, str):
                assert len(params) == 0, \
                    f"If the key is str it must be empty " \
                    f"(is not set) but it is: '{params}'"
                # empty param, skip it
                continue

            # grammar properties
            gramm = params.pop('gramm', '')
            if gramm:
                try:
                    gram_props = Corpus._parse_lexgramm_params(gramm, '|', True)
                except Exception:
                    raise
                self._params[f"gramm{word_num}"] = gram_props

            # additional properties
            flags = params.pop('flags', '')
            if flags:
                try:
                    flag_prop = Corpus._parse_lexgramm_params(flags, '+')
                except Exception:
                    raise
                self._params[f"flags{word_num}"] = flag_prop

            # TODO: semantic properties
            sem = params.pop('sem', '')
            if sem:
                logger.warning("Semantic properties does not support")

            if params:
                msg = f"Oops, 'gramm', 'flags' and 'sem' expected, but " \
                      f"another keys found: {params}"
                logger.error(msg)
                raise ValueError(msg)

    def _add_wordforms(self,
                       forms: List[str]) -> None:
        """ Add found wordforms to counter. """
        if not forms:
            return

        for form in forms:
            form = clean_text_up(form).lower()
            self._found_wordforms[form] += 1

    @abstractmethod
    def _parse_doc(self,
                   doc: bs4.element.Tag) -> Any:
        """ Parse the doc to list of Examples.

        Parsing depends on the subcorpus,
         the method redefined at the descendants.
        """
        # TODO: remake this func to generator?
        pass

    @abstractmethod
    def _parse_example(self,
                       *args,
                       **kwargs) -> Any:
        """ Parse the example to Example object.

        Parsing depends on the subcorpus,
         the method redefined at the descendants.
        """
        pass

    def _parse_kwic_example(self,
                            left: bs4.element.Tag,
                            center: bs4.element.Tag,
                            right: bs4.element.Tag) -> expl.KwicExample:
        l_txt = clean_text_up(left.text)
        c_txt = clean_text_up(center.text)
        # remove ←…→ symbol too
        r_txt = clean_text_up(right.text)[:-4].rstrip()

        found_wordforms = Corpus._find_searched_words(left)
        found_wordforms += Corpus._find_searched_words(center)
        found_wordforms += Corpus._find_searched_words(right)

        try:
            src = right.a.attrs['msg'].strip()
            url = right.a.attrs['href']
        except (KeyError, AttributeError, TypeError) as e:
            logger.error(f"Source or url not found:\n{e}")
            src = url = ''

        url = create_doc_url(url)

        new_ex = expl.KwicExample(
            l_txt, c_txt, r_txt, src, found_wordforms, url)
        new_ex.mark_found_words(self.marker)

        return new_ex

    def _parse_page_kwic(self,
                         page: str) -> List[expl.KwicExample]:
        """ Parse page if 'out' is 'kwic'.

        :exception ValueError: if the content not found.
        """
        soup = bs4.BeautifulSoup(page, 'lxml')
        res = []

        content = soup.find('table', {'align': 'left'})
        if not content:
            msg = "Content is None, this behavior " \
                  "is undefined, contact the developer"
            logger.critical(msg)
            raise ValueError(msg)

        nobr = content.find_all('nobr')
        if len(nobr) % 3:
            logger.warning("Len of nobr tags list % 3 != 0")

        for left, center, right in zip(nobr[::3], nobr[1::3], nobr[2::3]):
            new_ex = self._parse_kwic_example(left, center, right)
            res += [new_ex]
            self._add_wordforms(new_ex.found_wordforms)
        return res

    def _parse_page_normal(self,
                           page: str) -> List:
        """ Parse page if 'out' is 'normal'. """
        soup = bs4.BeautifulSoup(page, 'lxml')
        res = []

        for doc in soup.find_all('li'):
            try:
                parsed_doc = self._parse_doc(doc)
            except Exception as e:
                logger.error(f"Error while parsing doc:\n{e}")
            else:
                res += parsed_doc

        return res

    def _parse_all_pages(self,
                         pages: List[str]) -> List:
        """ Parse all pages. """
        parsed = [
            self._page_parser(page)
            for page in pages
        ]
        return sum(parsed, [])

    def _data_to_csv(self) -> None:
        """ Dump the data to csv file.
        Here it is assumed that the data exist.
        """
        data = [
            example.items
            for example in self.data
        ]
        columns = self[0].columns
        with self.file.open('w', encoding='utf-8', newline='') as f:
            # class constants
            dm = self._DATA_W_DELIMITER
            qch = self._DATA_W_QUOTCHAR

            writer = csv.writer(
                f, delimiter=dm, quotechar=qch, quoting=csv.QUOTE_MINIMAL)
            writer.writerows([columns] + data)

    def _params_to_json(self) -> None:
        """ Write the request params to json file.

        Here it is assumed that these params exist.
        """
        to_write = {
            'query': self.query,
            'p_count': self.p_count,
            'params': self.params
        }
        with self._config_path.open('w', encoding='utf-8') as f:
            ujson.dump(to_write, f, indent=4, ensure_ascii=False)

    def dump(self) -> None:
        """ Write the data to csv file, request params to json file.

        :return: None.
        :exception RuntimeError: If there are no data, params or files exist.
        """
        msg = "There is no data to write"
        if not self.data:
            logger.error(msg)
            raise RuntimeError(msg)
        if not (self.query and self.p_count and self.params):
            logger.error(msg)
            raise RuntimeError(msg)

        os.makedirs(self.DATA_FOLDER, exist_ok=True)

        self._data_to_csv()
        self._params_to_json()

        logger.info(
            f"Data wrote to files: {self.file} and {self._config_path}")

    def request_examples(self) -> None:
        """ Request examples, parse them and update the data.

        If there are no results found, last page does not exist,
        params or query is wrong then exception.

        :return: None.

        :exception RuntimeError: if the data still exist.
        """
        if self.data:
            logger.error("Tried to request new examples, however data exist")
            raise RuntimeError("Data still exist")

        start = time.time()
        try:
            first, last = creq.is_request_correct(
                RNC_URL, self.p_count, **self.params)
        except creq.BaseRequestError as e:
            msg = f"Query = {self.forms_in_query}, " \
                  f"{self.p_count}, {self.params}\ne = {e}"
            logger.error(msg)
            raise

        # get additional info from the first RNC page.
        logger.debug("Getting additional info from the first RNC page")
        if self.out == 'normal':
            self._get_additional_info(first)
        else:
            self._get_additional_info()
        logger.debug("Additional info received")

        if self.p_count > 2:
            logger.debug("Main request")
            htmls = creq.get_htmls(RNC_URL, 1, self.p_count - 1, **self.params)
            htmls = [first] + htmls + [last]
            logger.debug("Main request completed")
        else:
            htmls = [first]
            if self.p_count == 2:
                htmls += [last]

        logger.debug("Parsing html started")
        try:
            parsing_start = time.time()
            parsed = self._parse_all_pages(htmls)
            parsing_stop = time.time()
        except Exception as e:
            logger.error(f"Error while parsing, query = {self.params}\n{e}")
            raise
        else:
            logger.debug("Parsing completed")
            logger.info(f"Parsing time: {parsing_stop - parsing_start:.2f}")
            logger.info(f"Overall time: {parsing_stop - start:.2f}")
            self._data = parsed[:]

    async def request_examples_async(self) -> None:
        """ Request examples, parse them and update the data.

        If there are no results found, last page does not exist,
        params or query is wrong then exception.

        :return: None.

        :exception RuntimeError: if the data still exist.
        """
        if self.data:
            logger.error("Tried to request new examples, however data exist")
            raise RuntimeError("Data still exist")
        loop = asyncio.get_running_loop()

        start = time.time()
        try:
            first, last = await creq.is_request_correct_async(
                RNC_URL, self.p_count, **self.params)
        except creq.BaseRequestError as e:
            msg = f"Query = {self.forms_in_query}, " \
                  f"{self.p_count}, {self.params}\ne = {e}"
            logger.error(msg)
            raise

        # get additional info from the first RNC page.
        logger.debug("Getting additional info from the first RNC page")
        if self.out == 'normal':
            await self._get_additional_info_async(first)
        else:
            await self._get_additional_info_async()
        logger.debug("Additional info received")

        if self.p_count > 2:
            logger.debug("Main request")
            htmls = await creq.get_htmls_async(
                RNC_URL, 1, self.p_count - 1, **self.params)
            htmls = [first] + htmls + [last]
            logger.debug("Main request completed")
        else:
            htmls = [first]
            if self.p_count == 2:
                htmls += [last]

        logger.debug("Parsing html started")
        try:
            parsing_start = time.time()
            # sth wrong with ProcessPoolExecutor
            with ThreadPoolExecutor(os.cpu_count()) as executor:
                parsed = await loop.run_in_executor(
                    executor, self._parse_all_pages, htmls
                )
            parsing_stop = time.time()
        except Exception as e:
            logger.error(f"Error while parsing, query = {self.params}\n{e}")
            raise
        else:
            logger.debug("Parsing completed")
            logger.info(f"Parsing time: {parsing_stop - parsing_start:.2f}")
            logger.info(f"Overall time: {parsing_stop - start:.2f}")
            self._data = parsed[:]

    def copy(self) -> Any:
        copy_obj = self.__class__(
            self.query, self.p_count, file=self.file,
            marker=self.marker, **self.params)
        copy_obj._data = self.data.copy()
        return copy_obj

    def sort_data(self,
                  **kwargs) -> None:
        """ Sort the data by using a key.

        :keyword key: func to sort, called to Example objects, by default – len.
        :keyword reverse: bool, whether the data will sort in reversed order,
         by default – False.

        :exception TypeError: if the key is uncallable.
        """
        key = kwargs.pop('key', lambda example: len(example))
        reverse = kwargs.pop('reverse', False)

        if not callable(key):
            logger.error("Given uncallable key to sort")
            raise TypeError("Sort key must be callable")
        self._data.sort(key=key, reverse=reverse)

    def pop(self,
            index: int) -> Any:
        """ Remove and return element from data at the index. """
        return self._data.pop(index)

    def shuffle(self) -> None:
        """ Shuffle list of examples. """
        random.shuffle(self._data)

    def clear(self) -> None:
        """ Clear examples list. """
        self._data.clear()

    def filter(self,
               key: Callable) -> None:
        """ Remove some items, that are not satisfied the key.

        :param key: callable, it will be used to Example
        objects inside the data list.
        :return: None.
        """
        filtered_data = list(filter(key, self.data))
        self._data = filtered_data[:]

    def findall(self,
                pattern: Pattern or str,
                *args) -> Tuple[expl.Example, List[str]]:
        """ Apply the pattern to the examples' text with re.findall.
        Yield all examples which are satisfy the pattern and match.
        """
        for example in self:
            match = re.findall(pattern, example.txt, *args)
            if match:
                yield example, match

    def finditer(self,
                 pattern: Pattern or str,
                 *args) -> Tuple[expl.Example, Any]:
        """ Apply the pattern to the examples' text with re.finditer.
        Yield all examples which are satisfy the pattern and match.
        """
        for example in self:
            match = re.finditer(pattern, example.txt, *args)
            if match:
                yield example, match

    def __repr__(self) -> str:
        """ Format:
                Classname
                Length
                Database filename
                Request params
                Pages count
                Request
        """
        return f"{self.__class__.__name__}\n" \
               f"{len(self)}\n" \
               f"{self.file}\n" \
               f"{self.params}\n" \
               f"{self.p_count}\n" \
               f"{self.query}\n"

    def __str__(self) -> str:
        """
        :return: info about Corpus and enumerated examples.
        """
        q_forms = ', '.join(self.forms_in_query)
        metainfo = f"Russian National Corpus (https://ruscorpora.ru)\n" \
                   f"Class: {self.__class__.__name__}, len = {len(self)}\n" \
                   f"{self.p_count} pages of '{q_forms}' requested"

        data = self.data
        is_restricted = False
        if self.__RESTRICT_SHOW is not False and \
                len(data) > self.__RESTRICT_SHOW:
            data = self.data[:self.__RESTRICT_SHOW]
            is_restricted = True

        examples = '\n\n'.join(
            f"{num}.\n{str(example)}"
            for num, example in enumerate(data, 1)
        )
        examples += '\n...' * is_restricted

        return f"{metainfo}\n\n{examples}"

    def __len__(self) -> int:
        return len(self.data)

    def __bool__(self) -> bool:
        return bool(self.data)

    def __call__(self) -> None:
        """ All the same to request_examples() """
        self.request_examples()

    def __iter__(self) -> iter:
        return iter(self.data)

    def __contains__(self,
                     item: Any) -> bool:
        """ Whether the Corpus obj contains the Example obj.

        :param item: obj with the same ex_type.

        :exception TypeError: if wrong type (different Example) given.
        """
        if not isinstance(item, self.ex_type):
            msg = f"'in' must get the same Example " \
                  f"objects, but '{item.__class__.__name__}' found"
            logger.error(msg)
            raise TypeError(msg)
        return any(
            item == example
            for example in self.data
        )

    def __getattr__(self,
                    item: str) -> str or int or List or None:
        """ Get request param.

        :return: param value or None if it does not exist.
        """
        try:
            return getattr(super(), item)
        except AttributeError:
            return self.params.get(item, None)

    def __getitem__(self,
                    item: int or slice) -> Any:
        r""" Get example from data or create
         new obj with sliced data.

         Examples:
         =========
         .. code-block:: python
             >>> corp = MainCorpus(...)
             # get second example (1 is index, not number!)
             >>> corp[1]
             # create new copus with the first 50 example
             >>> new_corp = corp[:50]

        :return: one example or new obj with the same class and sliced data.
        :exception TypeError: if wrong type given.
        """
        if not isinstance(item, (int, slice)):
            msg = f"Int or slice expected, but '{item}' found"
            logger.error(msg)
            raise TypeError(msg)

        if isinstance(item, int):
            return self.data[item]

        new_data = self.data[item]
        new_obj = self.copy()
        new_obj._data = new_data.copy()
        return new_obj

    def __setitem__(self,
                    index: int,
                    new_example: Any) -> None:
        """ Change the example.

        Examples:
        >>> corp = MainCorpus(...)
        >>> corp[10] = MainExample(...)

        :exception TypeError: if wrong type given.
        """
        if not isinstance(new_example, self.ex_type):
            msg = f"{type(self.ex_type)} expected, " \
                  f"but {type(new_example)} found"
            logger.error(msg)
            raise TypeError(msg)

        try:
            self._data[index] = new_example
        except Exception as e:
            logger.error(f'Setting item: {new_example} to {index}\n{e}')
            raise

    def __delitem__(self,
                    key: int or slice) -> None:
        """ Delete example at the index or
         remove several ones using slice.

        Examples:
        >>> corp = MainCoprus(...)
        # delete forth example (3 is index, not number!)
        >>> del corp[3]
        # delete all examples after 10th
        >>> del corp[10:]
        # delete all exampes at even indexes from 0 to 10
        >>> del corp[0:10:2]

        :param key: int or slice, address of item(s) to delete.
        """
        try:
            del self._data[key]
        except Exception as e:
            logger.error(f"Deleting item: {key}\n{e}")
            raise


class MainCorpus(Corpus):
    _MODE = 'main'

    def __init__(self, *args, **kwargs):
        # for descendants
        ex_type = kwargs.pop('ex_type', expl.MainExample)
        super().__init__(*args, **kwargs, ex_type=ex_type)
        self._params['mode'] = self._MODE

    def _parse_example(self,
                       example: bs4.element.Tag):
        """ Parse example to Example object. """
        src = Corpus._get_source(example)
        txt = Corpus._get_text(example)
        txt = txt[:txt.index(src)]
        txt = txt[:txt.rindex('[')].strip()

        doc_url = Corpus._get_doc_url(example)
        ambiguation = Corpus._get_ambiguation(example)
        found_words = Corpus._find_searched_words(example)

        new_ex = self.ex_type(txt, src, ambiguation, found_words, doc_url)
        new_ex.mark_found_words(self.marker)
        return new_ex

    def _parse_doc(self,
                   doc: bs4.element.Tag) -> List[expl.MainExample]:
        """ Parse document to list of examples. """
        if not doc:
            logger.debug(f"Empty doc found, params: {self.params}")
            return []
        res = []

        for example in doc.find_all('li'):
            new_ex = self._parse_example(example)
            res += [new_ex]
            self._add_wordforms(new_ex.found_wordforms)
        return res


class NGrams(Corpus):
    # env = sas1_2
    pass


class BiGrams(NGrams):
    pass


class ThreeGrams(NGrams):
    pass


class FourGrams(NGrams):
    pass


class FiveGrams(NGrams):
    pass


class SyntaxCorpus(Corpus):
    pass


class Paper2000Corpus(MainCorpus):
    _MODE = 'paper'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, ex_type=expl.Paper2000Example)
        self._params['mode'] = self._MODE


class PaperRegionalCorpus(MainCorpus):
    _MODE = 'regional'

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs, ex_type=expl.PaperRegionalExample)
        self._params['mode'] = self._MODE


class ParallelCorpus(Corpus):
    _MODE = 'para'

    def __init__(self, *args, **kwargs) -> None:
        # for descendants
        ex_type = kwargs.pop('ex_type', expl.ParallelExample)
        super().__init__(*args, **kwargs, ex_type=ex_type)
        self._params['mode'] = self._MODE

    def _parse_text(self,
                    lang: str,
                    text: bs4.element.Tag) -> Any:
        """ Parse one element of the pair: original – translation.
        Means parse original or translation.
        """
        src = Corpus._get_source(text)
        txt = Corpus._get_text(text)
        # remove source from text
        txt = txt[:txt.index(src)]
        txt = txt[:txt.rindex('[')].strip()

        found_words = Corpus._find_searched_words(text)

        new_txt = self.ex_type(
            txt={lang: txt},
            src=src,
            ambiguation=Corpus._get_ambiguation(text),
            found_wordforms=found_words,
            doc_url=Corpus._get_doc_url(text)
        )
        new_txt.mark_found_words(self.marker)
        return new_txt

    def _parse_example(self,
                       tag: bs4.element.Tag) -> Any:
        """ Parse a pair: original – translation to Example. """
        # this example is expected to have default args
        result_example = self.ex_type()

        langs = tag.find_all('td', {'class': "para-lang"})
        texts = tag.find_all('li')
        for lang, text in zip(langs, texts):
            lang = lang.text.strip()
            new_txt = self._parse_text(lang, text)
            result_example += new_txt
        return result_example

    def _parse_doc(self,
                   doc: bs4.element.Tag) -> List:
        """ Parse one document. """
        res = []
        for example in doc.find_all('table', {'class': 'para'}):
            new_ex = self._parse_example(example)
            res += [new_ex]
            self._add_wordforms(new_ex.found_wordforms)
        return res

    def _load_data(self) -> List:
        """ Load data from csv file. """
        if self.out == 'kwic':
            return super()._load_data()

        with self.file.open('r', encoding='utf-8') as f:
            dm = self._DATA_W_DELIMITER
            qch = self._DATA_W_QUOTCHAR
            reader = csv.reader(f, delimiter=dm, quotechar=qch)

            columns = next(reader)
            end_lang_tags = columns.index('source')
            lang_tags = columns[:end_lang_tags]
            data = []

            for row in reader:
                # to create dict {lang: text in the lang}
                langs = {}
                for num, lang in enumerate(lang_tags):
                    langs[lang] = row[num]

                new_ex = self.ex_type(langs, *row[end_lang_tags:])
                data += [new_ex]

                self._add_wordforms(new_ex.found_wordforms)

        return data


class MultilingualParaCorpus(ParallelCorpus):
    _MODE = 'multi'

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs,
                         ex_type=expl.MultilingualParaExample)
        self._params['mode'] = self._MODE

    def _from_file(self) -> None:
        msg = f"Working with files not supported" \
              f" in {self.__class__.__name__}"
        logger.error(msg)
        raise NotImplementedError(msg)

    def dump(self) -> None:
        msg = f"Working with files not supported" \
              f" in {self.__class__.__name__}"
        logger.error(msg)
        raise NotImplementedError(msg)


class TutoringCorpus(MainCorpus):
    _MODE = 'school'

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs,
                         ex_type=expl.TutoringExample)
        self._params['mode'] = self._MODE


class DialectalCorpus(MainCorpus):
    _MODE = 'dialect'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs,
                         ex_type=expl.DialectalExample)
        self._params['mode'] = self._MODE


# save lines
class PoeticCorpus(Corpus):
    pass


class SpokenCorpus(MainCorpus):
    _MODE = 'spoken'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs,
                         ex_type=expl.SpokenExample)
        self._params['mode'] = self._MODE


class AccentologicalCorpus(MainCorpus):
    _MODE = 'accent'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs,
                         ex_type=expl.AccentologicalExample)
        self._params['mode'] = self._MODE


class MultimodalCorpus(Corpus):
    MEDIA_FOLDER = Corpus.DATA_FOLDER / 'media'
    _MODE = 'murco'

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs, ex_type=expl.MultimodalExample)
        self._params['mode'] = self._MODE

    def _parse_example(self,
                       example: bs4.element.Tag) -> Tuple[
        str, str, str, list, str]:
        """ Parse example get text, source etc. """
        src = Corpus._get_source(example)
        txt = Corpus._get_text(example)
        txt = txt[:txt.index(src)]
        txt = txt[:txt.rindex('[')].strip()

        doc_url = Corpus._get_doc_url(example)
        ambiguation = Corpus._get_ambiguation(example)
        found_words = Corpus._find_searched_words(example)

        return txt, src, ambiguation, found_words, doc_url

    def _parse_media(self,
                     media: bs4.element.Tag) -> Tuple[str, str]:
        """ Get link to the media file and filepath. """
        try:
            media_link = media.find('a')['href']
        except Exception:
            raise

        media_link, filename = media_link.split('?name=')
        return media_link, self.MEDIA_FOLDER / filename

    def _parse_doc(self,
                   doc: bs4.element.Tag) -> List[Any]:
        """ Parse the documents to examples. """
        try:
            media = doc.find('td', {'valign': 'top'})
            example = doc.find('td', {'class': 'murco-snippet'})
        except ValueError:
            return []
        examples = []

        media_url, filename = self._parse_media(media)
        # for example in example:
        data_from_example = self._parse_example(example)

        new_ex = self.ex_type(*data_from_example, media_url, filename)
        new_ex.mark_found_words(self.marker)
        self._add_wordforms(new_ex.found_wordforms)
        examples += [new_ex]

        return examples

    def download_all(self) -> None:
        """ Download all files. """
        os.makedirs(self.MEDIA_FOLDER, exist_ok=True)

        urls_to_names = [
            (example._media_url, example.filepath)
            for example in self
        ]
        creq.download_docs(urls_to_names)

    async def download_all_async(self) -> None:
        """ Download all files. """
        os.makedirs(self.MEDIA_FOLDER, exist_ok=True)

        urls_to_names = [
            (example._media_url, example.filepath)
            for example in self
        ]
        await creq.download_docs_async(urls_to_names)


class MultiPARCCorpus(Corpus):
    pass


class HistoricalCorpus(Corpus):
    pass



__version__ = '0.6.5'


import logging
import os
import sys
from pathlib import Path

from .corpora import (
    MainCorpus,
    Paper2000Corpus,
    PaperRegionalCorpus,
    ParallelCorpus,
    DialectalCorpus,
    SpokenCorpus,
    AccentologicalCorpus,
    MultilingualParaCorpus,
    TutoringCorpus,
    MultimodalCorpus,

    SORT_KEYS,
    OUTPUT_FORMATS,
    SEARCH_FORMATS
)
from .corpora_params import Mycorp
from .examples import (
    MainExample,
    Paper2000Example,
    PaperRegionalExample,
    ParallelExample,
    DialectalExample,
    SpokenExample,
    AccentologicalExample,
    MultilingualParaExample,
    TutoringExample,
    MultimodalExample,
    KwicExample
)

MSG_FMT = "[{asctime},{msecs:3.0f}] [{levelname}] " \
          "[{module}:{funcName}] {message}"
DATE_FMT = "%d.%m.%Y %H:%M:%S"

LOGGER_NAME = "rnc"
LOG_FOLDER = Path('logs')
LOG_FILE = LOG_FOLDER / f"{LOGGER_NAME}.log"
os.makedirs(LOG_FOLDER, exist_ok=True)


formatter = logging.Formatter(
    fmt=MSG_FMT, datefmt=DATE_FMT, style='{')

stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.WARNING)
stream_handler.setFormatter(formatter)

file_handler = logging.FileHandler(
    LOG_FILE, delay=True, encoding='utf-8')
file_handler.setLevel(logging.DEBUG)
file_handler.setFormatter(formatter)

logger = logging.getLogger(LOGGER_NAME)
logger.setLevel(logging.DEBUG)

logger.addHandler(stream_handler)
logger.addHandler(file_handler)


def set_handler_level(handler_class: type):
    def wrapped(level: int or str) -> None:
        try:
            level = level.upper()
        except AttributeError:
            pass

        for handler in logger.handlers:
            if isinstance(handler, handler_class):
                handler.setLevel(level)
                return
        print(f"There is no '{handler_class}' handler."
              f"This behavior is undefined, contact the developer",
              file=sys.stderr)

    return wrapped


set_stream_handler_level = set_handler_level(logging.StreamHandler)
set_file_handler_level = set_handler_level(logging.FileHandler)


def set_logger_level(level: int or str) -> None:
    try:
        level = level.upper()
    except AttributeError:
        pass
    logger.setLevel(level)


mycorp = Mycorp()

__all__ = (
    'MainCorpus',
    'Paper2000Corpus',
    'PaperRegionalCorpus',
    'ParallelCorpus',
    'DialectalCorpus',
    'SpokenCorpus',
    'AccentologicalCorpus',
    'MultilingualParaCorpus',
    'TutoringCorpus',
    'MultimodalCorpus',
    'mycorp',

    'MainExample',
    'Paper2000Example',
    'PaperRegionalExample',
    'ParallelExample',
    'DialectalExample',
    'SpokenExample',
    'AccentologicalExample',
    'MultilingualParaExample',
    'TutoringExample',
    'MultimodalExample',
    'KwicExample',

    'set_stream_handler_level',
    'set_file_handler_level',
    'set_logger_level',

    'SORT_KEYS',
    'SEARCH_FORMATS',
    'OUTPUT_FORMATS'
)



"""
Module for requesting to URL and get page's html code from there,
download media files, check that the request if correct, page in RNC exists.
"""

__all__ = (
    'get_htmls', 'is_request_correct', 'download_docs'
)

import asyncio
import logging
import time
from typing import List, Tuple

import aiofiles
import aiohttp
import bs4

logger = logging.getLogger("rnc")
WAIT = 24


class BaseRequestError(Exception):
    pass


class NoResultFound(BaseRequestError):
    pass


class LastPageDoesntExist(BaseRequestError):
    pass


class WrongHTTPRequest(BaseRequestError):
    pass


async def fetch_html(url: str,
                     ses: aiohttp.ClientSession,
                     **kwargs) -> Tuple[int, str] or None:
    """ Coro, obtaining page's HTML code.

    This coro should be awaited from a worker.

    :return: tuple of int and str, page index and its HTML code.
     None if there's an error, -1 if it's 429 and the worker should
     wait some time and make request again.

    :exception: all exceptions should be processed here.
    """
    worker_name = kwargs.pop('worker_name', '')
    try:
        resp = await ses.get(url, params=kwargs)
    except Exception as e:
        logger.error(
            f"{e}\n{worker_name}Cannot get "
            f"answer from '{url}' with {kwargs}")
        return

    if resp.status == 200:
        text = await resp.text('utf-8')
        resp.close()
        return kwargs['p'], text
    elif resp.status == 429:
        resp.close()
        return -1

    logger.error(
        f"{worker_name}{resp.status} -- '{resp.reason}' "
        f"requesting to {resp.url}"
    )
    resp.close()


async def worker_fetching_html(worker_name: str,
                               q_args: asyncio.Queue,
                               q_results: asyncio.Queue) -> None:
    """
    Worker requesting to URL with args from
     q_args and putting results to q_results.

    Wait some time and request again if there's 429 error.
    """
    while True:
        url, ses, kwargs = q_args.get_nowait()
        logger.debug(
            f"{worker_name}Requested to '{url}' with '{kwargs}'")

        res = await fetch_html(url, ses, **kwargs, worker_name=worker_name)

        if res is None:
            q_args.task_done()
            return

        while res == -1:
            logger.debug(
                    f"{worker_name}429 'Too many requests', "
                    f"page: {kwargs['p']}; wait {WAIT}s"
            )

            await asyncio.sleep(WAIT)
            res = await fetch_html(url, ses, **kwargs, worker_name=worker_name)

        logger.debug(
            f"{worker_name}Received from '{url}' with '{kwargs}'")
        q_args.task_done()

        await q_results.put((res[0], res[1]))


async def get_htmls_coro(url: str,
                         start: int,
                         stop: int,
                         **kwargs) -> List[str]:
    """
    Coro running 5 workers doing requests and
     getting HTML codes of the pages.

    URLs will be created for i in range(start, stop),
    HTTP tag 'p' (page) is i.

    """
    timeout = aiohttp.ClientTimeout(WAIT)

    q_results = asyncio.Queue(maxsize=-1)
    q_args = asyncio.Queue(maxsize=-1)

    async with aiohttp.ClientSession(timeout=timeout) as ses:
        for p_index in range(start, stop):
            await q_args.put((url, ses, {**kwargs, 'p': p_index}))

        tasks = []
        for worker_index in range(5):
            name = f"Worker-{worker_index + 1}: "
            task = asyncio.create_task(
                worker_fetching_html(name, q_args, q_results)
            )
            tasks += [task]

        await q_args.join()

        for task in tasks:
            task.cancel()

        results = [
            q_results.get_nowait()
            for _ in range(q_results.qsize())
        ]
    results.sort(key=lambda res: res[0])
    return [
        html for _, html in results
    ]


def get_htmls(url: str,
              start: int = 0,
              stop: int = 1,
              **kwargs) -> List[str]:
    """ Run coro, get html codes of the pages."""
    logger.info(f"Requested to '{url}' [{start};{stop}) with params {kwargs}")
    coro_start = time.time()

    html_codes = asyncio.run(
        get_htmls_coro(url, start, stop, **kwargs)
    )

    logger.info("Request was successfully completed")
    logger.info(f"Coro executing time: {round(time.time() - coro_start, 2)}")
    return html_codes


async def get_htmls_async(url: str,
                          start: int = 0,
                          stop: int = 1,
                          **kwargs) -> List[str]:
    """ Run coro, get html codes of the pages."""
    logger.info(f"Requested to '{url}' [{start};{stop}) with params {kwargs}")
    coro_start = time.time()

    html_codes = await get_htmls_coro(url, start, stop, **kwargs)

    logger.info("Request was successfully completed")
    logger.info(f"Coro executing time: {round(time.time() - coro_start, 2)}")
    return html_codes


def whether_result_found(url: str,
                         **kwargs) -> str:
    """
    Whether the page contains results.

    :return: first page HTML code if everything is OK.

    :exception RuntimeError: if HTTP request was wrong.
    :exception ValueError: if the result not found.
    """
    logger.debug("Validating that the request is OK")
    try:
        page_html = get_htmls(url, **kwargs)[0]
    except Exception:
        logger.error(f"The request is not correct: {kwargs}")
        raise RuntimeError
    logger.debug("The request is correct")

    logger.debug("Validating that the result exits")
    soup = bs4.BeautifulSoup(page_html, 'lxml')

    # TODO: сузить круг поиска
    content = soup.find('div', {'class': 'content'}).text
    res_msg = ('По этому запросу ничего не найдено.' in content or
               'No results match the search query.' in content)
    if res_msg:
        raise ValueError
    return page_html


def does_page_exist(url: str,
                    p_index: int,
                    first_page: str,
                    **kwargs) -> str:
    """
    Whether a page at the index exists.

    It means, the number of the page in 'pager' is equal to expected index.
    RNC redirects to the first page if the page at the number doesn't exist.
    Here it's assumed, that the request's correct.

    :return: last page code if everything is OK.

    :exception ValueError: the page doesn't exist.
    """
    # indexing starts with 0
    start = p_index
    start = start * (start >= 0)
    stop = p_index + 1

    # request's correct → first page exists
    if stop == 1:
        return first_page

    last_page = get_htmls(url, start, stop, **kwargs)[0]
    soup = bs4.BeautifulSoup(last_page, 'lxml')

    pager = soup.find('p', {'class': 'pager'})
    if pager:
        max_page_number = max(
            int(page.text)
            for page in pager.find_all('a')
            if page.text.isdigit()
        )
        if not max_page_number:
            raise ValueError
        if max_page_number < stop:
            raise ValueError
        return last_page

    # if there's no pager, but result exists.
    # this might happen if expand=full or out=kwic
    if last_page == first_page:
        raise ValueError
    return last_page


def is_request_correct(url: str,
                       p_count: int,
                       **kwargs) -> Tuple[str, str]:
    """
    Check:
        – is the HTTP request correct (means there are no exceptions catch).

        – has there been any result.

        – does a page at the number exist (
        means RNC doesn't redirect to the first page).

    :return: first and last pages if everything's OK.

    :exception WrongHTTPRequest: HTTP request is wrong.
    :exception NoResultFound: no result found.
    :exception LastPageDoesntExist: the last page doesn't exist.
    """
    logger.debug("Validating that everything is OK")
    try:
        # to reduce the number of requests
        # the two checks are combined into one.
        # coro writes logs by itself
        first_page = whether_result_found(url, **kwargs)
    except ValueError:
        logger.error("HTTP request is OK, but no result found")
        raise NoResultFound(f"{kwargs}")
    except RuntimeError:
        logger.error("HTTP request is wrong")
        raise WrongHTTPRequest(f"{kwargs}")
    logger.debug("HTTP request is correct, result found")

    logger.debug("Validating that the last page exists")
    try:
        last_page = does_page_exist(url, p_count - 1, first_page, **kwargs)
    except ValueError:
        logger.error("Everything is OK, but last page doesn't exist")
        raise LastPageDoesntExist(f"{kwargs}")
    logger.debug("The last page exists")

    logger.debug("Validated successfully")
    return first_page, last_page


async def whether_result_found_async(url: str,
                                     **kwargs) -> str:
    """
    Whether the page contains results.

    :return: first page HTML code if everything is OK.

    :exception RuntimeError: if HTTP request was wrong.
    :exception ValueError: if the result not found.
    """
    logger.debug("Validating that the request is OK")
    try:
        page_html = (await get_htmls_async(url, **kwargs))[0]
    except Exception:
        logger.error(f"The request is not correct: {kwargs}")
        raise RuntimeError
    logger.debug("The request is correct")

    logger.debug("Validating that the result exits")
    soup = bs4.BeautifulSoup(page_html, 'lxml')

    # TODO: сузить круг поиска
    content = soup.find('div', {'class': 'content'}).text
    res_msg = ('По этому запросу ничего не найдено.' in content or
               'No results match the search query.' in content)
    if res_msg:
        raise ValueError
    return page_html


async def does_page_exist_async(url: str,
                                p_index: int,
                                first_page: str,
                                **kwargs) -> str:
    """
    Whether a page at the index exists.

    It means, the number of the page in 'pager' is equal to expected index.
    RNC redirects to the first page if the page at the number doesn't exist.
    Here it's assumed, that the request's correct.

    :return: last page code if everything is OK.

    :exception ValueError: the page doesn't exist.
    """
    # indexing starts with 0
    start = p_index
    start = start * (start >= 0)
    stop = p_index + 1

    # request's correct → first page exists
    if stop == 1:
        return first_page

    last_page = (await get_htmls_async(url, start, stop, **kwargs))[0]
    soup = bs4.BeautifulSoup(last_page, 'lxml')

    pager = soup.find('p', {'class': 'pager'})
    if pager:
        max_page_number = max(
            int(page.text)
            for page in pager.find_all('a')
            if page.text.isdigit()
        )
        if not max_page_number:
            raise ValueError
        if max_page_number < stop:
            raise ValueError
        return last_page

    # if there's no pager, but result exists.
    # this might happen if expand=full or out=kwic
    if last_page == first_page:
        raise ValueError
    return last_page


async def is_request_correct_async(url: str,
                                   p_count: int,
                                   **kwargs) -> Tuple[str, str]:
    """
    Check:
        – is the HTTP request correct (means there are no exceptions catch).

        – has there been any result.

        – does a page at the number exist (
        means RNC doesn't redirect to the first page).

    :return: first and last pages if everything's OK.

    :exception WrongHTTPRequest: HTTP request is wrong.
    :exception NoResultFound: no result found.
    :exception LastPageDoesntExist: the last page doesn't exist.
    """
    logger.debug("Validating that everything is OK")
    try:
        # to reduce the number of requests
        # the two checks are combined into one.
        # coro writes logs by itself
        first_page = await whether_result_found_async(url, **kwargs)
    except ValueError:
        logger.error("HTTP request is OK, but no result found")
        raise NoResultFound(f"{kwargs}")
    except RuntimeError:
        logger.error("HTTP request is wrong")
        raise WrongHTTPRequest(f"{kwargs}")
    logger.debug("HTTP request is correct, result found")

    logger.debug("Validating that the last page exists")
    try:
        last_page = await does_page_exist_async(url, p_count - 1, first_page, **kwargs)
    except ValueError:
        logger.error("Everything is OK, but last page doesn't exist")
        raise LastPageDoesntExist(f"{kwargs}")
    logger.debug("The last page exists")

    logger.debug("Validated successfully")
    return first_page, last_page


async def fetch_media_file(url: str,
                           ses: aiohttp.ClientSession,
                           **kwargs) -> bytes or int:
    """
    Coro, getting media content to write.

    :return: bytes (media) if everything is OK,
     -1 if there's 429 error, None if it is another error.

    :exception: all exceptions should be processed here.
    """
    worker_name = kwargs.pop('worker_name', '')
    try:
        resp = await ses.get(url, allow_redirects=True, params=kwargs)
    except Exception as e:
        logger.error(
            f"{e}\n{worker_name}Cannot get "
            f"answer from '{url}' with {kwargs}")
        return

    if resp.status == 200:
        content = await resp.read()
        resp.close()
        return content
    elif resp.status == 429:
        resp.close()
        return -1

    logger.error(
        f"{resp.status}: {resp.reason} requesting to {resp.url}"
    )
    resp.close()


async def dump(content: bytes,
               filename: str) -> None:
    """ Dump content to media file."""
    async with aiofiles.open(filename, 'wb') as f:
        await f.write(content)


async def worker_fetching_media(worker_name: str,
                                q_args: asyncio.Queue) -> None:
    """
    Worker getting media file and dumping it to file.

    Wait some time and request again if there's 429 error.
    """
    while True:
        url, ses, filename = q_args.get_nowait()

        logger.debug(f"{worker_name}Requested to '{url}'")
        content = await fetch_media_file(url, ses, worker_name=worker_name)

        if content is None:
            q_args.task_done()
            return

        while content == -1:
            logger.debug(
                f"{worker_name}: 429 'Too many requests', "
                f"url: {url}; wait {WAIT}s"
            )

            await asyncio.sleep(WAIT)
            content = await fetch_media_file(url, ses, worker_name=worker_name)

        logger.debug(f"{worker_name}Received from '{url}'")
        logger.debug(f"{worker_name}Dumping '{url}' to '{filename}'")
        await dump(content, filename)
        logger.debug(f"{worker_name}'{filename}' dumped")

        q_args.task_done()


async def download_docs_coro(url_to_name: List[Tuple[str, str]]) -> None:
    """ Coro running 5 workers to download media files. """
    timeout = aiohttp.ClientTimeout(WAIT)
    q_args = asyncio.Queue(maxsize=-1)

    async with aiohttp.ClientSession(timeout=timeout) as ses:
        for url, filename in url_to_name:
            await q_args.put((url, ses, filename))

        tasks = []
        for worker_number in range(5):
            name = f"Worker-{worker_number + 1}: "
            task = asyncio.create_task(
                worker_fetching_media(name, q_args))
            tasks += [task]

        await q_args.join()

        for task in tasks:
            task.cancel()


def download_docs(url_to_name: List[Tuple[str, str]]) -> None:
    """
    Run coro, download the files.

    :param url_to_name: list of tuples of str, pairs: url – filename.
    """
    logger.info(f"Requested {len(url_to_name)} files to download")
    coro_start = time.time()

    asyncio.run(download_docs_coro(url_to_name))

    logger.info(f"Downloading completed, coro executing time: "
                f"{round(time.time() - coro_start, 2)}s")


async def download_docs_async(url_to_name: List[Tuple[str, str]]) -> None:
    """
    Run coro, download the files.

    :param url_to_name: list of tuples of str, pairs: url – filename.
    """
    logger.info(f"Requested {len(url_to_name)} files to download")
    coro_start = time.time()

    await download_docs_coro(url_to_name)

    logger.info(f"Downloading completed, coro executing time: "
                f"{round(time.time() - coro_start, 2)}s")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rnc']

package_data = \
{'': ['*']}

install_requires = \
['aiofiles>=0.7.0,<0.8.0',
 'aiohttp>=3.7.4.post0,<3.8.0',
 'beautifulsoup4>=4.9.3,<4.10.0',
 'bs4>=0.0.1,<0.1.0',
 'lxml>=4.6.3,<4.7.0',
 'ujson>=4.1.0,<4.2.0']

setup_kwargs = {
    'name': 'rnc',
    'version': '0.7.0',
    'description': 'API for Russian National Corpus',
    'long_description': "# API for [Russian National Corpus](http://ruscorpora.ru) \n\n## Installation\n```bash\npip install rnc\n```\n\n## Structure\nCorpus object contains list of obtained examples.\nThere are two types of examples:\n![](https://github.com/kunansy/RNC/blob/master/docs/Two_ex_types.png?raw=true) <br> \n* If `out` is `normal`, API uses normal example, which name is equal to the Corpus class name:\n\n```python\nru = rnc.MainCorpus(...)\nru.request_examples()\n\nprint(type(ru[0]))\n>>> MainExample\n```\n* if `out` is `kwic`, API uses `KwicExample`.\n\nExamples' objects [fields](https://github.com/kunansy/RNC/blob/master/docs/Examples.md)   \n\n\n## Usage\n```python\nimport rnc\n\nru = rnc.MainCorpus(\n    query='корпус', \n    p_count=5,\n    file='filename.csv',\n    marker=str.upper,\n    **kwargs\n)\n\nru.request_examples()\n```\n* `query` – one str or dict with tags. Words to find, you should give the vocabulary form of them.\n* `p_count` – count of **PAGES**.\n* `file` – path to local csv file, optional. Example: `file='data\\\\filename.csv'`. \n* `marker` – function, with which found wordforms will be marked, optional. \n* `kwargs` – additional params.\n\n[Corpora](https://github.com/kunansy/RNC/blob/master/docs/Corpora.md) you can use.\n\n\n### Full query form\n```python\nquery = {\n    'word1': {\n        'gramm': 'acc', # grammar tags for lexgramm search\n        'flags': 'bdot' # additional tags for lexgramm search\n    },\n    # you can get as a value one string or dict of params\n    # params are: any name of dict key, name of tag (you can see them below)  \n    'word2': {\n        'gramm': { \n            # the NAMES of these keys might be any\n            'pos (any name)': 'S' or ['S', 'A'], # one value or list of values,\n            'case (any name)': 'acc' or ['acc', 'nom'],\n        },\n        'flags': {}, # all the same to here\n        # distance between first and second words\n        'min': 1,  \n        'max': 3\n    },  \n}\n\ncorp = rnc.MainCorpus(\n    query, 5, file='filename.csv', marker=str.upper, **kwargs)\ncorp.reques_examples()\n```\n[Lexgramm search params](https://github.com/kunansy/RNC/tree/master/docs/Lexgram%20search%20params)\n\n\n### String as a query\nAlso you can pass as a query a string with the **vocabulary forms** of the \nwords, divided by space: `query = 'get down'` or `query = 'я получить'`. \nDistance between them will be default.\n\n\n### Additional request params\nThese params are optional, you can ignore them. Here are the default values.\n```python\ncorp = rnc.ParallelCorpus(\n    query=query, \n    p_count=5,\n    file='filename.csv',\n    marker=str.upper,\n    \n    dpp=5, # documents per page\n    spd=10, # sentences per document (<= than spd)\n    text='lexgramm' or 'lexform', # way to search\n    out='normal' or 'kwic', # output format\n    kwsz=5, # if out=kwic, count of words in context\n    sort='i_grtagging', # way to sort the results, see HOWTO section below\n    mycorp='', # see HOWTO section below\n    accent=0, # with accentology (1) or without (0), if it is available\n)\n```\n[Sort keys](https://github.com/kunansy/RNC/blob/master/docs/HTTP%20params.md)\n\n\n### API can work with a local file too\n```python\nru = rnc.SpokenCorpus(file='local_database.csv') # it must exist\nprint(ru)\n```\nIf the file exists, API works with it. If the data list is not empty you \ncannot request new examples. <br>\n\nIf you work with a file, it is not demanded to pass any argument to Corpus \nexcept for the file name (`file=...`).\n\n\n### Working with corpora\n```python\ncorp = rnc.corpus_name(...) \n```\n* `corp.request_examples()` – request examples. \nThere is an exception if:\n    * Data still exist. \n    * No results found.\n    * A requested page does not exist (if there are 10 pages in the RNC, but \n      you have requested > 10).\n    * There is a mistake in the request.\n    * You have no access to the Internet.\n    * There is a problem while getting access to RNC.\n    * another problems...\n* `corp.data` – list of examples (only getter)\n* `corp.query` – query (only getter).\n* `corp.forms_in_query` – requested wordforms (only getter).\n* `corp.p_count` – requested count of pages (only getter). \n* `corp.file` – path to the local csv file (only getter).\n* `corp.marker` – marker (only getter).\n* `corp.params` – dict, HTTP tags (only getter). \n* `corp.found_wordforms` – dict with found wordforms and their frequency (only getter).\n* `corp.ex_type` – type of example (only getter).\n* `corp.amount_of_docs` – amount of docs where the query was found.\n* `corp.amount_of_contexts` – amount of contexts where the query was found.\n* `corp.graphic_link` – link to the graphic of the distribution of query occurrences by years.\n* `corp.dump()` – write two files: csv file with all data and json file with config.\n* `corp.copy()` – create a copy.\n* `corp.shuffle()` – shuffle data list.\n* `corp.sort_data(key=, reverse=)` – sort the list of examples. Here HTTP keys do not work,\nkey is applied to Example objects.  \n* `corp.pop(index)` – remove and return the example at the index.\n* `corp.clear()` – empty the data list.\n* `corp.filter(key)` – filter the data list, remove some examples using the key. \nKey is applied to the `Example` objects.\n* `corp.url` – URL of the first RNC page (only getter).\n* `corp.findall(pattern, args)` – get all examples where the pattern found and \n  the match.\n* `corp.finditer(pattern, args)` – get all examples where the pattern found and \n  the match.\n\nMagic methods: \n* `corp.dpp` or another request param (only getter).\n* `corp()` – all the same to `request_examples()`.\n* `str(corp) or print(corp)` – str with info about Corpus, enumerated examples.\nBy default, Corpus shows first 50 examples, but you can change it \nor turn the restriction off. \n\n    Info about Corpus:\n    ```\n    Russian National Corpus (https://ruscorpora.ru)\n    Class: CorpusName, len = amount of examples \n    Pages: n of 'words' requested\n    ```\n* `len(corp)` – count of examples.\n* `bool(corp)` – whether data exist.\n* `corp[index or slice]` – get element at the index or create a new object \n  with sliced data:\n```python\nfrom_2_to_10 = corp[2:10:2]\n```\n* `del corp[10]` or `del corp[:10]` – remove some examples from the data list.\n\n* Also you can use cycle `for`. For example we want to see only left \n  context (`out=kwic`) and source:\n```python\ncorp = rnc.ParallelCorpus(\n    'corpus', 5, \n    out='kwic', kwsz=7, \n    mycorp=rnc.mycorp.en\n)\ncorp.request_examples()\n\nfor r in corp:\n    print(r.left)\n    print(r.src)\n```\n\nSet default values to all objects you will create:\n* `corpus_name.set_dpp(value)` – change default `document per page` value.\n* `corpus_name.set_spd(value)` – change default `sentences per document` value.\n* `corpus_name.set_text(value)` – change default search way.\n* `corpus_name.set_sort(value)` – change default sort key.\n* `corpus_name.set_min(value)` – change default min distance between words.\n* `corpus_name.set_max(value)` – change default max distance between words.\n* `corpus_name.set_restrict_show(value)` – change default amount of shown examples in print. \nIf it is equal to `False`, the Corpus shows all examples. \n\n\n### Corpora features\n#### ParallelCorpus\n* The query might be both in the original language and in the language of \n  translation. \n\n#### MultilingualParaCorpus\n* Working with files is removed.\n* Param `mycorp` is not demanded by default, but it might be passed, see \n  **HOWTO** section below.\n\n#### MultimodalCorpus\n* `corp.download_all()` – download all media files. **It is recommended** to use \nthis method instead of `expl.download_file()`.\n\n\n## Logger\n* See all log messages\n```python\nrnc.set_stream_handler_level('debug')\n```\n* See less than all messages\n```python\nrnc.set_stream_handler_level('info')\n```\n* Turn the logger off\n```python\nrnc.set_logger_level('critical')\n```\n* Turn off all messages in the stream, but dump logs to file\n```python\nrnc.set_stream_handler_level('critical')\n```\n* Turn off dumping logs to file\n```python\nrnc.set_file_handler_level('critical')\n```\n\n\n## ATTENTION\n* Do not forget to call this function\n```python\ncorp.request_examples()\n```\n* If you have requested more than 10 pages, RNC returns 429 error \n  (Too many requests).\nFor example requesting 100 pages you should wait about 3 minutes: \n![100 pages](https://github.com/kunansy/RNC/blob/master/docs/100_pages.png?raw=true)\n* **Do not call** the marker you pass\n\n**RIGHT:**\n```python\nru = rnc.MainCorpus(...,  marker=str.upper)\n```\n**WRONG:**\n```python\nru = rnc.MainCorpus(..., marker=str.upper())\n```\n* Pass an empty string as a param if you do not want to set them\n```python\nquery = {\n    'word1': '',\n    'word2': {'min': 2, 'max': 5}\n}\n```\n* If `accent=1`, marker does not work.\n\n---\n\n## HOWTO\nYou can ask any question you want [here](https://github.com/kunansy/RNC/discussions).\n\n### How to set sort?\nThere are some sort keys:\n1. `i_grtagging` – by default.\n2. `random` – randomly.\n3. `i_grauthor` – by author.\n4. `i_grcreated_inv` – by creation date.\n5. `i_grcreated` – by creation date in reversed order.\n6. `i_grbirthday_inv` – by author's birth date.\n7. `i_grbirthday` – by author's birth date in reversed order.\n\n[Some of HTTP params](https://github.com/kunansy/RNC/blob/master/docs/HTTP%20params.md).\n\n\n### How to set language in ParallelCorpus?\n```python\nen = rnc.ParallelCorpus('get', 5, mycorp=rnc.mycorp.en)\n```\n**OR**\n```python\nen = rnc.ParallelCorpus('get', 5, mycorp=rnc.mycorp['en'])\n```\nLanguage keys list:\n1. Armenian – 'arm'\n1. Bashkir – 'bas'\n1. Belarusian – 'bel'\n1. Bulgarian – 'bul'\n1. Buryatian – 'bur'\n1. Chinese – 'ch'\n1. Czech – 'cz'\n1. English – 'en'\n1. Estonian – 'es'\n1. Finnish – 'fin'\n1. French – 'fr'\n1. German – 'ger'\n1. Italian – 'it'\n1. Latvian – 'lat'\n1. Lithuanian – 'lit'\n1. Polish – 'pol'\n1. Spanish – 'sp'\n1. Swedish – 'sw'\n1. Ukrainian – 'ukr'\n\nIf you want to search something by several languages, choose and set the \n`mycorp` in the site, pass this param to Corpus. \n\n\n### How to set subcorpus?\nMeans specify the sample where you want to search the query. <br>\n\nThere are default keys in `rnc.mycorp` (working checked in \n**MainCorpus**) – Russian writers and poets: \n* Pushkin\n* Dostoyevsky\n* TolstoyLN\n* Chekhov\n* Gogol\n* Turgenev\n\nExample:\n```python\nru = rnc.MainCorpus('нету', 1, mycorp=rnc.mycorp['Pushkin'])\n```\n\n\n**OR**\n\n```python\nru = rnc.MainCorpus('нету', 1, mycorp=rnc.mycorp.Pushkin)\n```\n\n\n**OR**\n\n \n![1](https://raw.githubusercontent.com/kunansy/RNC/master/docs/How%20to%20set%20subcorpus/1.png)\n![2](https://raw.githubusercontent.com/kunansy/RNC/master/docs/How%20to%20set%20subcorpus/2.png)\n![3](https://raw.githubusercontent.com/kunansy/RNC/master/docs/How%20to%20set%20subcorpus/3.png)\n![4](https://raw.githubusercontent.com/kunansy/RNC/master/docs/How%20to%20set%20subcorpus/4.png)\n\n\n## Links\n* [Russian National Corpus](https://ruscorpora.ru)\n* [Docs](https://github.com/kunansy/RNC/tree/master/docs)\n* Examples' objects [fields](https://github.com/kunansy/RNC/blob/master/docs/Examples.md)\n* [Corpora](https://github.com/kunansy/RNC/blob/master/docs/Corpora.md) you can use.\n* [Lexgramm search params](https://github.com/kunansy/RNC/tree/master/docs/Lexgram%20search%20params)\n* [Sort keys](https://github.com/kunansy/RNC/blob/master/docs/HTTP%20params.md)\n---\n\n\n## Requirements\n* Python >= 3.7\n\n\n## Licence\n`rnc` is offered under MIT licence.\n\n\n## Source code\nThe project is hosted on [Github](https://github.com/kunansy/RNC)\n\n---\n\nPlease file an issue in the [bug tracker](https://github.com/kunansy/RNC/issues) \nif you have found a bug or have some suggestions to improve the library.\n",
    'author': 'kunansy',
    'author_email': 'kolobov.kirill@list.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://kunansy.github.io/RNC/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
